# Generated by Django 5.2 on 2025-08-25 12:53
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "ipfabric_netbox",
            "0017_ipfabricsync_update_custom_fields",
        ),
    ]
    operations = [
        migrations.RemoveField(
            model_name="ipfabricsync",
            name="type",
        ),
    ]
